import numpy as np
import pandas as pd


def swap(img):
    img_swap = img.swapaxes(0, img.ndim - 1)
    return img_swap


def stack_to_list(stack):
    num_ims = stack.shape[0]
    return [stack[im, :] for im in range(num_ims)]


def list_to_stack(list):
    return np.array(list)


def max_proj(stack):
    im = np.max(stack, axis=0)
    return im


def scale(stack, metric, clip=False):
    from starfish.stats import stack_describe
    stats = stack_describe(stack)
    ims = stack_to_list(stack)
    res = [im / s[metric] for im, s in zip(ims, stats)]
    return list_to_stack(res)


def gather(df, key, value, cols):
    id_vars = [col for col in df.columns if col not in cols]
    id_values = cols
    var_name = key
    value_name = value
    return pd.melt(df, id_vars, id_values, var_name, value_name)


def spots_to_geojson(spots_viz):
    '''
    Convert spot geometrical data to geojson format
    '''

    def make_dict(row):
        row = row[1]
        d = dict()
        d['properties'] = {'id': int(row.spot_id), 'radius': int(row.r)}
        d['geometry'] = {'type': 'Point', 'coordinates': [int(row.x), int(row.y)]}
        return d

    return [make_dict(row) for row in spots_viz.iterrows()]


def regions_to_geojson(r):
    '''
    Convert region geometrical data to geojson format
    '''

    def make_dict(id, verts):
        d = dict()
        c = list(map(lambda x: list(x), list(map(lambda v: [int(v[0]), int(v[1])], verts))))
        d["properties"] = {"id": id}
        d["geometry"] = {"type": "Polygon", "coordinates": c}
        return d

    return [make_dict(id, verts) for id, verts in enumerate(r.hull)]


def relabel(image):
    '''
    This is a local implementation of centrosome.cpmorphology.relabel
    to remove this dependency from starfish.

    Takes a labelled image and relabels each image object consecutively.
    Original code from:
    https://github.com/CellProfiler/centrosome/blob/master/centrosome/cpmorphology.py

    They use a BSD-3 license, which would then have to be propagated to starfish,
    this could be an issue.

    Args
    ----
    image: numpy.ndarray
        A 2d integer array representation of an image with labels

    Returns
    -------
    new_image: numpy.ndarray
        A 2d integer array representation of an image wiht new labels

    n_labels: int
        The number of new unique labels
    '''

    # I've set this as a separate function, rather than binding it to the
    # WatershedSegmenter object for now

    unique_labels = set(image[image != 0])
    n_labels = len(unique_labels)

    # if the image is unlabelled, return original image
    # warning/message required?
    if n_labels == 0:
        return (image, 0)

    consec_labels = np.arange(n_labels) + 1
    lab_table = np.zeros(max(unique_labels) + 1, int)
    lab_table[[x for x in unique_labels]] = consec_labels

    # Use the label table to remap all of the labels
    new_image = lab_table[image]

    return new_image, n_labels



from __future__ import division

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from showit import image
from skimage.feature import blob_log

from starfish.munge import stack_to_list, gather


class GaussianSpotDetector:
    def __init__(self, stack, blobs, measurement_type='max'):
        self.stack = stack
        self.blobs = blobs
        self.measurement_type = measurement_type

        self.num_objs = None
        self.spots_df = None
        self.spots_df_viz = None
        self.intensities = None

    def detect(self, min_sigma, max_sigma, num_sigma, threshold):
        fitted_blobs = self.fit(min_sigma, max_sigma, num_sigma, threshold)
        spots_df_viz = self._fitted_blobs_to_df(fitted_blobs)
        intensity = self._measure(self.blobs, spots_df_viz)
        spots_df_viz['intensity'] = intensity
        spots_df_viz['spot_id'] = spots_df_viz.index
        self.spots_df_viz = spots_df_viz
        self.intensities = self._measure_stack()

    def fit(self, min_sigma, max_sigma, num_sigma, threshold):
        fitted_blobs = blob_log(self.blobs, min_sigma, max_sigma, num_sigma, threshold)
        fitted_blobs[:, 2] = fitted_blobs[:, 2] * np.sqrt(2)
        self.num_objs = fitted_blobs.shape[0]
        return fitted_blobs

    @staticmethod
    def _fitted_blobs_to_df(fitted_blobs):
        res = pd.DataFrame(fitted_blobs)
        res['d'] = res[2] * 2
        res.rename(columns={0: 'x', 1: 'y', 2: 'r'}, inplace=True)
        res['xmin'] = np.floor(res.x - res.r)
        res['xmax'] = np.ceil(res.x + res.r)
        res['ymin'] = np.floor(res.y - res.r)
        res['ymax'] = np.ceil(res.y + res.r)
        return res

    def _measure(self, img, spots_df):
        res = []
        for row in spots_df.iterrows():
            row = row[1]
            subset = img[int(row.xmin):int(row.xmax), int(row.ymin):int(row.ymax)]
            if self.measurement_type == 'max':
                try:
                    res.append(subset.max())
                except:
                    res.append(np.NAN)
            else:
                try:
                    res.append(subset.mean())
                except:
                    res.append(np.NAN)
        return res

    def _measure_stack(self):
        intensities = [self._measure(img, self.spots_df_viz) for img in stack_to_list(self.stack)]
        return intensities

    def to_encoder_dataframe(self, tidy_flag, mapping=None):
        inds = range(self.stack.shape[0])
        d = dict(zip(inds, self.intensities))
        d['spot_id'] = range(self.num_objs)

        res = pd.DataFrame(d)

        if tidy_flag:
            res = gather(res, 'ind', 'val', inds)
            if mapping is not None:
                res = pd.merge(res, mapping, on='ind', how='left')
                del res['ind']

        self.spots_df = res

        return self.spots_df

    def to_viz_dataframe(self):
        return self.spots_df_viz

    def show(self, figsize=(20, 20)):
        plt.figure(figsize=figsize)
        ax = plt.gca()

        image(self.blobs, ax=ax)
        blobs_log = self.spots_df_viz.loc[:, ['x', 'y', 'r']].values
        for blob in blobs_log:
            x, y, r = blob
            c = plt.Circle((y, x), r, color='r', linewidth=2, fill=False)
            ax.add_patch(c)

        plt.title('Num blobs: {}'.format(len(blobs_log)))
        plt.show()



import numpy as np
import pandas as pd
from skimage.measure import points_in_poly

from .stats import label_to_regions


def assign(cells_label, spots, use_hull=True, verbose=False):
    cells_region = label_to_regions(cells_label)

    res = pd.DataFrame({'spot_id': range(0, spots.shape[0])})
    res['cell_id'] = None

    for cell_id in range(cells_region.count):
        if use_hull:
            verts = cells_region[cell_id].hull
        else:
            verts = cells_region[cell_id].coordinates
        verts = np.array(verts)
        in_poly = points_in_poly(spots, verts)
        res.loc[res.spot_id[in_poly], 'cell_id'] = cell_id
        if verbose:
            cnt = np.sum(in_poly)
            print(cell_id, cnt)

    return res



from __future__ import division

import numpy as np
import pandas as pd
from skimage.filters import threshold_otsu
import scipy.ndimage.measurements as spm
from starfish.stats import label_to_regions


class PixelSpotDetector:
    def __init__(self, stack, blobs):
        self.stack = stack
        self.blobs = blobs

        self.num_objs = None
        self.spots_df = None
        self.spots_df_viz = None
        self.threshold = None

    def detect(self):
        self.spots_df_viz = self.to_viz_dataframe()
        self.spots_df = self.to_encoder_dataframe()

    def to_encoder_dataframe(self):
        """
        Writes table of spot_id | hyb | ch | val
        :param min_intensity: minimum intensity of pixel value to retain 
        :return: spots_df 
        """

        def encode(row, squeezed_stack):
            r = row[1]
            sq = squeezed_stack[r.ind]
            sqt = sq[self.threshold[:, 0], self.threshold[:, 1]]
            val = sqt.ravel()
            d = pd.DataFrame()
            d['spot_id'] = range(self.num_objs)
            d['val'] = val
            d['hyb'] = r.hyb
            d['ch'] = r.ch
            return d

        squeezed_stack = self.stack.squeeze()
        squeeze_map = self.stack.squeeze_map
        res = [encode(row, squeezed_stack) for row in squeeze_map.iterrows()]
        self.spots_df = pd.concat(res)

        return self.spots_df

    def to_viz_dataframe(self):
        """
        Determines pixel locations worthy of decoding. Uses otsu thresholding to separate
        signal from background. This decreases search space necessary for pixel based readout
        :return: spots_df_viz: | spot_id | x | y | z | 
        """

        # determine threshold
        t = threshold_otsu(self.blobs)
        blobs_binary = self.blobs > t
        ind = np.argwhere(blobs_binary)
        self.threshold = ind

        # construct spots_viz
        self.num_objs = ind.shape[0]
        df = pd.DataFrame()
        df['spot_id'] = range(self.num_objs)
        df['x'] = ind[:, 0]
        df['y'] = ind[:, 1]

        # assign groupings to spots (pixels)
        labels, num_objs = spm.label(blobs_binary)
        regions = label_to_regions(labels)
        for grp_num, r in enumerate(regions):
            cond_1 = df.x.isin(r.coordinates[:, 0])
            cond_2 = df.y.isin(r.coordinates[:, 1])
            cond = cond_1 & cond_2
            df.loc[cond, 'spot_group'] = grp_num

        self.spots_df_viz = df

        return self.spots_df_viz



import numpy as np
import pandas as pd


def decode(spots_df_tidy, letters=['T', 'G', 'C', 'A']):
    num_ch = spots_df_tidy.ch.max() + 1
    num_hy = spots_df_tidy.hyb.max() + 1
    num_spots = spots_df_tidy.spot_id.max() + 1

    seq_res = np.zeros((num_spots, num_hy))
    seq_stren = np.zeros((num_spots, num_hy))
    seq_qual = np.zeros((num_spots, num_hy))

    for spot_id in range(num_spots):

        sid_df = spots_df_tidy[spots_df_tidy.spot_id == spot_id]

        mat = np.zeros((num_hy, num_ch))
        inds = zip(sid_df.hyb.values, sid_df.ch.values, sid_df.val)

        for tup in inds:
            mat[tup[0], tup[1]] = tup[2]

        max_stren = np.max(mat, axis=1)
        max_ind = np.argmax(mat, axis=1)
        qual = max_stren / np.sum(mat, axis=1)

        seq_res[spot_id, :] = max_ind
        seq_stren[spot_id, :] = max_stren
        seq_qual[spot_id, :] = qual

    max_qual = np.max(seq_qual, axis=1)

    codes = []
    for k in range(seq_res.shape[0]):
        letter_inds = seq_res[k, :]
        letter_inds = letter_inds.astype(np.int)
        res = ''.join([letters[j] for j in letter_inds])
        codes.append(res)

    dec = pd.DataFrame({'spot_id': range(num_spots),
                        'barcode': codes,
                        'qual': max_qual})

    return dec



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.ndimage.measurements as spm
from showit import image

from starfish.filters import bin_thresh
from starfish.munge import gather
from starfish.stats import label_to_regions, measure_stack


class BinarySpotDetector:
    def __init__(self, stack, thresh, blobs):
        self.stack = stack
        self.thresh = thresh

        self.blobs = blobs
        self.blobs_binary = None
        self.labels = None
        self.num_objs = None
        self.areas = None
        self.intensities = None

        self.regions = None
        self.spots_df = None

    def detect(self, measurement_type='mean'):
        self.blobs_binary = self._threshold()
        self.labels, self.num_objs = self._label()
        self.areas, self.intensities = self._measure(measurement_type)
        self.regions = self._to_regions()
        return self

    def _threshold(self):
        blobs_binary = bin_thresh(self.blobs, self.thresh)
        return blobs_binary

    def _label(self):
        labels, num_objs = spm.label(self.blobs_binary)
        return labels, num_objs

    def _measure(self, measurement_type):
        areas = spm.sum(np.ones(self.labels.shape),
                        self.labels,
                        range(1, self.num_objs))

        intensity = measure_stack(self.stack, self.labels, self.num_objs, measurement_type)
        return areas, intensity

    def _to_regions(self):
        regions = label_to_regions(self.labels)[1:]
        return regions

    def to_encoder_dataframe(self, tidy_flag):
        num_hybs = self.stack.shape[0]
        cols = range(num_hybs)
        cols = ['hyb_{}'.format(c + 1) for c in cols]
        d = dict(zip(cols, self.intensities))
        d['spot_id'] = range(self.num_objs - 1)

        res = pd.DataFrame(d)

        if tidy_flag:
            res = gather(res, 'hybs', 'vals', cols)

        self.spots_df = res

        return self.spots_df

    def to_viz_dataframe(self):
        res = pd.DataFrame(self.regions.center)
        res = res.rename(columns=dict(zip(res.columns, ['x', 'y', 'z'])))

        if 'z' not in res.columns:
            res['z'] = None
        res['spot_id'] = res.index
        res['area'] = self.areas
        return res

    def show(self, figsize=(10, 10)):
        plt.figure(figsize=figsize)
        plt.subplot(121)
        image(self.blobs_binary, size=10, ax=plt.gca())

        plt.subplot(122)
        regions = self.regions
        image(regions.mask(background=[0.9, 0.9, 0.9],
                           dims=self.labels.shape,
                           stroke=None,
                           cmap='rainbow'),
              size=10,
              ax=plt.gca()
              )



import pandas as pd
from skimage import io
import numpy as np

from .munge import list_to_stack
import json
import os


class Stack:
    def __init__(self):

        # data organization
        self.org = None
        self.path = None
        self.format = None

        # numpy array (num_hybs, num_chans, x, y, z)
        self.data = None

        # auxilary images
        self.aux_dict = dict()

        # shape data
        self.num_hybs = None
        self.num_chs = None
        self.im_shape = None
        self.is_volume = None
        self.squeeze_map = None

        # readers and writers
        self.read_fn = None  # set by self._read_metadata
        self.write_fn = np.save  # asserted for now

    @property
    def shape(self):
        if self.data is None:
            return None
        else:
            return self.data.shape

    def read(self, in_json):
        with open(in_json, 'r') as in_file:
            self.org = json.load(in_file)

        self.path = os.path.dirname(os.path.abspath(in_json)) + '/'
        self._read_metadata()
        self._read_stack()
        self._read_aux()

    def _read_metadata(self):
        d = self.org['metadata']
        self.num_hybs = d['num_hybs']
        self.num_chs = d['num_chs']
        self.im_shape = tuple(d['shape'])

        self.is_volume = d['is_volume']
        if not self.is_volume:
            self.data = np.zeros((self.num_hybs, self.num_chs, self.im_shape[0], self.im_shape[1]))
        else:
            self.data = np.zeros((self.num_hybs, self.num_chs, self.im_shape[0], self.im_shape[1], self.im_shape[2]))

        self.format = d['format']

        if self.format == 'tiff':
            self.read_fn = io.imread
        else:
            self.read_fn = np.load

    def _read_stack(self):
        data_dicts = self.org['data']

        for d in data_dicts:
            h = d['hyb']
            c = d['ch']
            fname = d['file']
            im = self.read_fn(os.path.join(self.path, fname))
            self.data[h, c, :] = im

    def _read_aux(self):
        data_dicts = self.org['aux']

        for d in data_dicts:
            typ = d['type']
            fname = d['file']
            self.aux_dict[typ] = self.read_fn(os.path.join(self.path, fname))

    # TODO should this thing write npy?
    def write(self, dir_name):
        self._write_metadata(dir_name)
        self._write_stack(dir_name)
        self._write_aux(dir_name)

    def _write_metadata(self, dir_name):
        new_org = self.org
        new_org['metadata']['format'] = 'npy'

        def format(d):
            d['file'] = self._swap_ext(d['file']) + '.npy'
            return d

        new_org['data'] = [format(d) for d in new_org['data']]
        new_org['aux'] = [format(d) for d in new_org['aux']]

        with open(os.path.join(dir_name, 'org.json'), 'w') as outfile:
            json.dump(new_org, outfile, indent=4)

        self.org = new_org

    def _write_stack(self, dir_name):
        for d in self.org['data']:
            h = d['hyb']
            c = d['ch']
            fname = d['file']
            self.write_fn(os.path.join(dir_name, fname), self.data[h, c, :])

    def _write_aux(self, dir_name):
        for d in self.org['aux']:
            typ = d['type']
            fname = d['file']
            self.write_fn(os.path.join(dir_name, fname), self.aux_dict[typ])

    def _swap_ext(self, fname):
        if self.format == 'tiff':
            fname = fname.replace('.tiff', '')
        else:
            fname = fname.replace('.npy', '')
        return fname

    def set_stack(self, new_stack):
        if new_stack.shape != self.shape:
            msg = "Shape mismatch. Current data shape: {}, new data shape: {}".format(self.shape, new_stack.shape)
            raise AttributeError(msg)
        self.data = new_stack

    def set_aux(self, key, img):
        if key in self.aux_dict:
            old_stack = self.aux_dict[key]
            if old_stack.shape != img.shape:
                msg = "Shape mismatch. Current data shape: {}, new data shape: {}".format(old_stack.shape,
                                                                                          img.shape)
                raise AttributeError(msg)
        else:
            self.org['aux'].append({'file': key, 'type': key})
        self.aux_dict[key] = img

    def max_proj(self, dim):
        valid_dims = ['hyb', 'ch', 'z']
        if dim not in valid_dims:
            msg = "Dimension: {} not supported. Expecting one of: {}".format(dim, valid_dims)
            raise ValueError(msg)

        if dim == 'hyb':
            res = np.max(self.data, axis=0)
        elif dim == 'ch':
            res = np.max(self.data, axis=1)
        elif dim == 'z' and self.is_volume:
            res = np.max(self.data, axis=4)
        else:
            res = self.data

        return res

    def squeeze(self, bit_map_flag=False):
        new_shape = ((self.num_hybs * self.num_chs),) + self.im_shape
        new_data = np.zeros(new_shape)

        # TODO this can all probably be done smartly with np.reshape instead of a double for loop
        ind = 0
        inds = []
        hybs = []
        chs = []

        for h in range(self.num_hybs):
            for c in range(self.num_chs):
                new_data[ind, :] = self.data[h, c, :]
                inds.append(ind)
                hybs.append(h)
                chs.append(c)
                ind += 1

        self.squeeze_map = pd.DataFrame({'ind': inds, 'hyb': hybs, 'ch': chs})

        if bit_map_flag:
            mp = [(d['hyb'], d['ch'], d['bit']) for d in self.org['data']]
            mp = pd.DataFrame(mp, columns=['hyb', 'ch', 'bit'])
            self.squeeze_map = pd.merge(self.squeeze_map, mp, on=['ch', 'hyb'], how='left')

        return new_data

    def un_squeeze(self, stack):
        if type(stack) is list:
            stack = list_to_stack(stack)

        new_shape = (self.num_hybs, self.num_chs) + self.im_shape
        res = np.zeros(new_shape)

        # TODO this can probably done smartly without a double for loop
        ind = 0
        for h in range(self.num_hybs):
            for c in range(self.num_chs):
                res[h, c, :] = stack[ind, :]
                ind += 1

        return res



import numpy as np
from scipy.ndimage import fourier_shift
from skimage.feature import register_translation

from .munge import stack_to_list, list_to_stack


def compute_shift(im, ref, upsample_factor=1):
    shift, error, diffphase = register_translation(im, ref, upsample_factor)
    return shift, error


def shift_im(im, shift):
    fim_shift = fourier_shift(np.fft.fftn(im), map(lambda x: -x, shift))
    im_shift = np.fft.ifftn(fim_shift)
    return im_shift.real


def register(im, ref, upsample_factor=1):
    shift, _ = compute_shift(im, ref, upsample_factor)
    res = shift_im(im, shift)
    return res


def register_stack(stack, ref, upsample_factor=1):
    ims = stack_to_list(stack)
    res = [register(i, ref, upsample_factor) for i in ims]
    return list_to_stack(res)



import numpy as np
import pandas as pd
from functools import reduce
import scipy.ndimage.measurements as spm
from regional import many as Many
from regional import one as One
from scipy.sparse import coo_matrix


def stack_describe(stack):
    num_hybs = stack.shape[0]
    stats = [im_describe(stack[k, :]) for k in range(num_hybs)]
    return stats


def im_describe(im):
    shape = im.shape
    flat_dims = reduce(lambda x, y: x * y, shape)
    flat_im = np.reshape(im, flat_dims)
    stats = pd.Series(flat_im).describe()
    return stats.to_dict()


def label_to_regions(labels):
    label_mat_coo = coo_matrix(labels)

    def region_for(label_mat_coo, label):
        ind = label_mat_coo.data == label
        # TODO does this work in 3D?
        x = label_mat_coo.row[ind]
        y = label_mat_coo.col[ind]
        # LOL -- in python3 zip returns an iterator. to force it to
        # a list we call list(zip). In Python 2.7 this is effectively a noop
        re = One(list(zip(x, y)))
        return re

    unique_labels = sorted(set(label_mat_coo.data))
    regions = [region_for(label_mat_coo, label) for label in unique_labels]

    return Many(regions)


def measure(im, labels, num_objs, measurement_type='mean'):
    if measurement_type == 'mean':
        res = spm.mean(im, labels, range(1, num_objs))
    elif measurement_type == 'max':
        res = spm.maximum(im, labels, range(1, num_objs))
    else:
        raise ValueError('Unsporrted measurement type: {}'.format(measurement_type))

    return res


def measure_stack(stack, labels, num_objs, measurement_type='mean'):
    from starfish.munge import stack_to_list
    ims = stack_to_list(stack)
    res = [measure(im, labels, num_objs, measurement_type) for im in ims]
    return res



import numpy as np
from showit import image, tile

from .stats import im_describe, stack_describe


def tile_lims(stack, num_std, bar=True, size=20):
    if type(stack) is list:
        stack = np.array(stack)
    stats = stack_describe(stack)
    lims = [s['mean'] + num_std * s['std'] for s in stats]
    tile(stack, bar=bar, clim=list(zip([0] * len(lims), lims)), size=size)


def image_lims(im, num_std, bar=True, size=20):
    stats = im_describe(im)
    lim = stats['mean'] + num_std * stats['std']
    image(im, bar=bar, size=size, clim=[0, lim])



import cv2
import numpy as np
from skimage import restoration
from skimage.filters import gaussian
from skimage.morphology import binary_erosion, binary_dilation, disk, binary_opening, binary_closing
from scipy.ndimage.filters import maximum_filter, minimum_filter

from .munge import swap, stack_to_list, list_to_stack


def gaussian_low_pass(img, sigma, ksize=None, border=None, skimage=False):
    img_swap = swap(img)
    if ksize is None:
        ksize = int(2 * np.ceil(2 * sigma) + 1)

    if border is None:
        border = cv2.BORDER_REPLICATE

    if not skimage:
        blurred = cv2.GaussianBlur(img_swap,
                                   (ksize, ksize),
                                   sigma,
                                   borderType=border
                                   )
    else:

        blurred = gaussian(img_swap,
                           sigma=sigma,
                           output=None,
                           cval=0,
                           multichannel=True,
                           preserve_range=True,
                           truncate=4.0
                           )

    blurred = blurred.astype(np.uint16)

    return swap(blurred)


def gaussian_high_pass(img, sigma, ksize=None, border=None, skimage=False):
    blurred = gaussian_low_pass(img, sigma, ksize, border, skimage)

    over_flow_ind = img < blurred
    res = img - blurred
    res[over_flow_ind] = 0

    return res


def richardson_lucy_deconv(img, num_iter, psf=None, gpar=None, clip=False):
    if psf is None:
        if gpar is None:
            msg = 'Must specify a gaussian (kernel size, sigma) if a psf is not specified'
            raise ValueError(msg)
        ksize, sigma = gpar
        psf = cv2.getGaussianKernel(ksize, sigma, cv2.CV_32F)
        psf = np.dot(psf, psf.T)

    img_deconv = restoration.richardson_lucy(img, psf, iterations=num_iter, clip=clip)

    # here be dragons. img_deconv is a float. this should not work, but the result looks nice
    # modulo boundary values? wtf indeed.
    img_deconv = img_deconv.astype(np.uint16)
    return img_deconv


def bin_erode(im, disk_size):
    selem = disk(disk_size)
    res = binary_erosion(im, selem)
    return res


def bin_dilate(im, disk_size):
    selem = disk(disk_size)
    res = binary_dilation(im, selem)
    return res


def bin_open(im, disk_size):
    selem = disk(disk_size)
    res = binary_opening(im, selem)
    return res


def bin_close(im, disk_size):
    selem = disk(disk_size)
    res = binary_closing(im, selem)
    return res


def bin_thresh(im, thresh):
    res = im >= thresh
    return res


def white_top_hat(im, disk_size):
    selem = disk(disk_size)
    min_filt = minimum_filter(im, footprint=selem)
    max_filt = maximum_filter(min_filt, footprint=selem)
    res = im - max_filt
    return res



import setuptools

setuptools.setup(
    name="starfish",
    version="0.0.0a",
    description="Pipelines and pipeline components for the analysis of image-based transcriptomics data",
    author="Deep Ganguli",
    author_email="dganguli@chanzuckerberg.com",
    license="MIT",
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': "starfish=starfish.starfish:starfish"
    }
)



from __future__ import division

import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage.measurements as spm
from scipy.ndimage import distance_transform_edt
from showit import image
from skimage.feature import peak_local_max
from skimage.morphology import watershed

from .filters import bin_thresh, bin_open
from .munge import relabel
from .stats import label_to_regions


class WatershedSegmenter:
    def __init__(self, dapi_img, stain_img):
        self.dapi = dapi_img / dapi_img.max()
        self.stain = stain_img / stain_img.max()

        self.dapi_thresholded = None
        self.markers = None
        self.num_cells = None
        self.mask = None
        self.segmented = None

    def segment(self, dapi_thresh, stain_thresh, size_lim, disk_size_markers=None, disk_size_mask=None, min_dist=None):
        min_allowed_size, max_allowed_size = size_lim
        self.dapi_thresholded = self.filter_dapi(dapi_thresh, disk_size_markers)
        self.markers, self.num_cells = self.label_nuclei(min_allowed_size, max_allowed_size, min_dist)
        self.mask = self.watershed_mask(stain_thresh, self.markers, disk_size_mask)
        self.segmented = self.watershed(self.markers, self.mask)
        return self.segmented

    def filter_dapi(self, dapi_thresh, disk_size):
        dapi_filt = bin_thresh(self.dapi, dapi_thresh)
        if disk_size is not None:
            dapi_filt = bin_open(dapi_filt, disk_size)
        return dapi_filt

    def label_nuclei(self, min_allowed_size, max_allowed_size, min_dist=None):

        if min_dist is None:
            markers, num_objs = spm.label(self.dapi_thresholded)
        else:
            markers, num_objs = self._unclump(min_dist)

        min_allowed_area = min_allowed_size ** 2
        max_allowed_area = max_allowed_size ** 2

        areas = spm.sum(np.ones(self.dapi_thresholded.shape),
                        markers,
                        np.array(range(0, num_objs + 1), dtype=np.int32))

        area_image = areas[markers]

        markers[area_image <= min_allowed_area] = 0
        markers[area_image >= max_allowed_area] = 0

        markers_reduced, num_objs = relabel(markers)

        return markers_reduced, num_objs

    def _unclump(self, min_dist):
        im = self.dapi_thresholded
        distance = distance_transform_edt(im)
        local_maxi = peak_local_max(distance, labels=im, indices=False, min_distance=min_dist)
        markers, num_objs = spm.label(local_maxi)
        labels_ws = watershed(-distance, markers, mask=im)
        return labels_ws, num_objs

    def watershed_mask(self, stain_thresh, markers, disk_size):
        st = self.stain >= stain_thresh
        watershed_mask = np.logical_or(st, markers > 0)
        if disk_size is not None:
            watershed_mask = bin_open(watershed_mask, disk_size)
        return watershed_mask

    def watershed(self, markers, watershed_mask):
        img = 1 - self.stain

        res = watershed(image=img,
                        markers=markers,
                        connectivity=np.ones((3, 3), bool),
                        mask=watershed_mask
                        )

        return res

    def to_regions(self):
        regions = label_to_regions(self.segmented)
        return regions

    def show(self, figsize=(10, 10)):
        plt.figure(figsize=figsize)

        plt.subplot(321)
        image(self.dapi, ax=plt.gca(), size=20, bar=True)
        plt.title('DAPI')

        plt.subplot(322)
        image(self.stain, ax=plt.gca(), size=20, bar=True)
        plt.title('Stain')

        plt.subplot(323)
        image(self.dapi_thresholded, bar=True, ax=plt.gca())
        plt.title('DAPI Thresholded')

        plt.subplot(324)
        image(self.mask, bar=True, ax=plt.gca())
        plt.title('Watershed Mask')

        plt.subplot(325)
        marker_regions = label_to_regions(self.markers)
        im = marker_regions.mask(background=[0.9, 0.9, 0.9], dims=self.markers.shape, stroke=None, cmap='rainbow')
        image(im, size=20, ax=plt.gca())
        plt.title('Found: {} cells'.format(self.num_cells))

        plt.subplot(326)
        segmented_regions = label_to_regions(self.segmented)
        im = segmented_regions.mask(background=[0.9, 0.9, 0.9], dims=self.segmented.shape, stroke=None, cmap='rainbow')
        image(im, size=20, ax=plt.gca())
        plt.title('Segmented Cells')

        return plt.gca()



#!/usr/bin/env python
import os

import click
import matplotlib.pyplot as plt
import numpy as np
import json
from showit import tile

from .assign import assign
from .filters import white_top_hat
from .io import Stack
from .register import compute_shift, shift_im
from .spots.gaussian import GaussianSpotDetector
from .watershedsegmenter import WatershedSegmenter
from .stats import label_to_regions
from .munge import spots_to_geojson, regions_to_geojson

import pandas as pd


@click.group()
def starfish():
    art = """
         _              __ _     _
        | |            / _(_)   | |
     ___| |_ __ _ _ __| |_ _ ___| |__
    / __| __/ _` | '__|  _| / __| '_  `
    \__ \ || (_| | |  | | | \__ \ | | |
    |___/\__\__,_|_|  |_| |_|___/_| |_|

    """
    print(art)


@starfish.command()
@click.argument('in_json', type=click.Path(exists=True))
@click.argument('out_dir', type=click.Path(exists=True))
@click.option('--u', default=1, help='Amount of up-sampling', type=int)
def register(in_json, out_dir, u):
    print('Registering ...')
    s = Stack()
    s.read(in_json)

    mp = s.max_proj('ch')
    res = np.zeros(s.shape)

    for h in range(s.num_hybs):
        # compute shift between maximum projection (across channels) and dots, for each hyb round
        shift, error = compute_shift(mp[h, :, :], s.aux_dict['dots'], u)
        print("For hyb: {}, Shift: {}, Error: {}".format(h, shift, error))

        for c in range(s.num_chs):
            # apply shift to all channels and hyb rounds
            res[h, c, :] = shift_im(s.data[h, c, :], shift)

    s.set_stack(res)

    s.write(out_dir)


@starfish.command()
@click.argument('in_json', type=click.Path(exists=True))
@click.argument('out_dir', type=click.Path(exists=True))
@click.option('--ds', default=15, help='Disk size', type=int)
def filter(in_json, out_dir, ds):
    print('Filtering ...')
    print('Reading data')
    s = Stack()
    s.read(in_json)

    # filter raw images, for all hybs and channels
    stack_filt = []
    for im_num, im in enumerate(s.squeeze()):
        print("Filtering image: {}...".format(im_num))
        im_filt = white_top_hat(im, ds)
        stack_filt.append(im_filt)

    stack_filt = s.un_squeeze(stack_filt)

    # filter dots
    print("Filtering dots ...")
    dots_filt = white_top_hat(s.aux_dict['dots'], ds)

    print("Writing results ...")
    # create a 'stain' for segmentation
    stain = np.mean(s.max_proj('ch'), axis=0)
    stain = stain / stain.max()

    # update stack
    s.set_stack(stack_filt)
    s.set_aux('dots', dots_filt)
    s.set_aux('stain', stain)

    s.write(out_dir)


@starfish.command()
@click.argument('in_json', type=click.Path(exists=True))
@click.argument('results_dir', type=click.Path(exists=True))
@click.argument('aux_img', type=str)
@click.option('--min_sigma', default=4, help='Minimum spot size (in standard deviation)', type=int)
@click.option('--max_sigma', default=6, help='Maximum spot size (in standard deviation)', type=int)
@click.option('--num_sigma', default=20, help='Number of scales to try', type=int)
@click.option('--t', default=.01, help='Dots threshold', type=float)
@click.option('--show', default=False, help='Dots threshold', type=bool)
def detect_spots(in_json, results_dir, aux_img, min_sigma, max_sigma, num_sigma, t, show):
    print('Finding spots...')
    s = Stack()
    s.read(in_json)

    gsp = GaussianSpotDetector(s.squeeze(), s.aux_dict[aux_img])

    gsp.detect(min_sigma, max_sigma, num_sigma, t)
    if show:
        gsp.show(figsize=(10, 10))

    spots_viz = gsp.spots_df_viz
    spots_df_tidy = gsp.to_encoder_dataframe(tidy_flag=True, mapping=s.squeeze_map)

    geojson = spots_to_geojson(spots_viz)

    path = os.path.join(results_dir, 'spots.json')
    print("Writing | spots geojson to: {}".format(path))
    with open(path, 'w') as f:
        f.write(json.dumps(geojson))

    path = os.path.join(results_dir, 'spots_geo.csv')
    print("Writing | spot_id | x | y | z | to: {}".format(path))
    spots_viz.to_csv(path, index=False)

    path = os.path.join(results_dir, 'encoder_table.csv')
    print("Writing | spot_id | hyb | ch | val | to: {}".format(path))
    spots_df_tidy.to_csv(path, index=False)


@starfish.command()
@click.argument('in_json', type=click.Path(exists=True))
@click.argument('results_dir', type=click.Path(exists=True))
@click.argument('aux_image')
@click.option('--dt', default=.16, help='DAPI threshold', type=float)
@click.option('--st', default=.22, help='Input threshold', type=float)
@click.option('--md', default=57, help='Minimum distance between cells', type=int)
def segment(in_json, results_dir, aux_image, dt, st, md):
    s = Stack()
    s.read(in_json)

    # TODO make these parameterizable or determine whether they are useful or not
    size_lim = (10, 10000)
    disk_size_markers = None
    disk_size_mask = None

    seg = WatershedSegmenter(s.aux_dict['dapi'], s.aux_dict[aux_image])
    cells_labels = seg.segment(dt, st, size_lim, disk_size_markers, disk_size_mask, md)

    r = label_to_regions(cells_labels)
    geojson = regions_to_geojson(r)

    path = os.path.join(results_dir, 'regions.json')
    print("Writing | regions geojson to: {}".format(path))
    with open(path, 'w') as f:
        f.write(json.dumps(geojson))

    spots_geo = pd.read_csv(os.path.join(results_dir, 'spots_geo.csv'))
    # TODO only works in 3D
    points = spots_geo.loc[:, ['x', 'y']].values
    res = assign(cells_labels, points, use_hull=True)

    path = os.path.join(results_dir, 'regions.csv')
    print("Writing | cell_id | spot_id to: {}".format(path))
    res.to_csv(path, index=False)


@starfish.command()
@click.argument('results_dir', type=click.Path(exists=True))
@click.option('--decoder_type', default='iss', help='Decoder type')
def decode(results_dir, decoder_type):
    if decoder_type == 'iss':
        from .decoders.iss import decode as dec
    else:
        raise ValueError('Decoder type: {} not supported'.format(decoder_type))

    encoder_table = pd.read_csv(os.path.join(results_dir, 'encoder_table.csv'))
    res = dec(encoder_table)
    path = os.path.join(results_dir, 'decoder_table.csv')
    print("Writing | spot_id | gene_id to: {}".format(path))
    res.to_csv(path, index=False)


@starfish.command()
@click.argument('in_json', type=click.Path(exists=True))
@click.option('--sz', default=10, help='Figure size', type=int)
def show(in_json, sz):
    s = Stack()
    s.read(in_json)
    tile(s.squeeze(), size=sz, bar=True)
    plt.show()
